# Copyright (C) 2022 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Any
from typing import Callable
from typing import Generator


class CollectionMixin:
    """Mixin class for collection types."""
    __module__: str = 'canonical'
    openapi_title: str | None = None
    openapi_format: str | None = None
    max_length: int | None = None
    min_length: int | None = None

    @classmethod
    def __modify_schema__(
        cls,
        field_schema: dict[str, Any]
    ) -> None:
        field_schema.update( # pragma: no cover
            title=cls.openapi_title or cls.__name__,
            type='array'
        )
        if cls.openapi_format:
            field_schema.update(format=cls.openapi_format)

    @classmethod
    def __get_validators__(cls) -> Generator[Callable[..., Any], None, None]:
        yield cls.validate

    @classmethod
    def validate(cls: type[Any], v: Any) -> Any:
        return cls(v)


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from datetime import date
from datetime import datetime

import pydantic

from .shipmenttrackingnumber import ShipmentTrackingNumber


class ShipmentTrackingInfo(pydantic.BaseModel):
    """Contains information to track the status of a :term:`Shipment` consisting
    of one or multiple packages.
    """
    carrier_id: str = pydantic.Field(
        default=...,
        title="Carrier ID",
        description="An identifier for the the Carrier, usually a domain name."
    )

    carrier_name: str = pydantic.Field(
        default=...,
        title="Carrier",
        description=(
            "The human-readable name of the carrier that is providing the "
            "shipping services."
        )
    )

    trace: ShipmentTrackingNumber | None = pydantic.Field(
        default=None,
        title="Tracking information",
        description=(
            "Tracking information regarding the shipment and its items."
        )
    )

    etd: datetime | date = pydantic.Field(
        default=...,
        title="Estimated Time of Departure (ETD)",
        description=(
            "Expected Time of Departure is the prediction of time "
            "that is expected for a transport system to depart "
            "its point of origin or location. As shipping labels "
            "are usually created prior to actual shipment, this "
            "estimation should be interpreted as *on or before*."
        )
    )
    
    def add(self, trackingcode: str, trackingurl: str | None) -> None:
        if self.trace is not None:
            raise ValueError("ShipmentTrackingInfo.tracking_number is already set.")
        self.trace = ShipmentTrackingNumber(
            number=trackingcode,
            url=trackingurl
        )


# Copyright (C) 2022 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import pytest
import pytz

from canonical import TimezoneInfo



@pytest.mark.parametrize("timezone",
    [x for x in pytz.all_timezones if '/' in x] # type: ignore
)
def test_verify_valid_zone(timezone: str):
    TimezoneInfo.validate(timezone)


def test_unknown_zone_raises_valueerror():
    with pytest.raises(ValueError):
        TimezoneInfo.validate('foo')


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import re
from typing import cast
from typing import TypeVar

from .domainname import DomainName
from .stringtype import StringType


T = TypeVar('T', bound='ResourceName')


class ResourceName(StringType):
    relname: str
    service: DomainName
    patterns: list[re.Pattern[str]] = [
        re.compile(r'//.*')
    ]

    def __new__(
        cls: type[T],
        object: str,
        service: DomainName | None = None,
        relname: str | None = None
    ) -> T:
        self = super().__new__(cls, object)
        if service is None:
            # Was not created by pydantic
            for validator in cls.__get_validators__():
                self = validator(self)
        else:
            assert relname is not None
            self.relname = relname
            self.service = service
        return cast(T, self)
    
    @classmethod
    def validate(cls, v: str) -> str:
        assert str.startswith(v, '//')
        service, _, relname = str.partition(v[2:], '/')
        if not relname:
            raise ValueError("a valid ResourceName contains a relative name.")
        for validator in DomainName.__get_validators__():
            service = validator(service)
        return cls(v, relname=relname, service=DomainName(service))


# Copyright (C) 2022 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Any
from typing import Callable
from typing import Generator

import pytz
from pydantic.validators import str_validator


TIMEZONES: list[str] = [x for x in pytz.all_timezones if '/' in x] # type: ignore


class TimezoneInfo(str):

    @classmethod
    def __modify_schema__(
        cls,
        field_schema: dict[str, Any]
    ) -> None:
        field_schema.update( # pragma: no cover
            title='Timezone',
            type='string',
            format='string'
        )

    @classmethod
    def __get_validators__(cls) -> Generator[Callable[..., str | None], None, None]:
        yield str_validator
        yield cls.validate

    @classmethod
    def validate(cls, v: Any) -> str:
        if len(v) > 128:
            raise ValueError("Value is too long to be a valid time zone.")
        if v not in TIMEZONES:
            raise ValueError(f"Not a valid time zone: {v}")
        return cls(v)

    def __repr__(self) -> str: # pragma: no cover
        return f'EmailAddress({self})'


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from ..collectionmixin import CollectionMixin
from .shipmenttrackinginfo import ShipmentTrackingInfo


class ShipmentTrackingInfoList(list[ShipmentTrackingInfo], CollectionMixin):

    def has(self, trackingcode: str) -> bool:
        """Return a boolean if the :class:`ShipmentTrackingInfoList`
        holds any shipment with the given tracking code.
        """
        return any([getattr(x.trace, 'number', None) is not None for x in self])